module Scriptroute
module Tulip

############# whether a given IP speaks a particular language ##########
module LangChecker;

  def LangChecker.getBestOption(destination, *prefList)
    prefList.each { |type| 
      case type
      when "tstamp"
        return "tstamp" if(LangChecker.probeResponse(destination, 
                                                     Scriptroute::ICMP::ICMP_TSTAMP, 
                                                     Scriptroute::ICMP::ICMP_TSTAMPREPLY)) 
      when "echo"
        return "echo" if(LangChecker.probeResponse(destination, 
                                                   Scriptroute::ICMP::ICMP_ECHO, 
                                                   Scriptroute::ICMP::ICMP_ECHOREPLY))
      when "udp"
        return "udp" if(LangChecker.udp(destination))
      when "tcp" 
        return "tcp" if(LangChecker.tcp(destination))
      else
        raise("ERROR: unknown packet type = #{type}");
      end
    }
    return nil;
  end

  def LangChecker.getBestOption4IPIDs(destination, *prefList)
    prefList.each { |type| 
      return [type, true] if(LangChecker.increasingIPIDs(destination, 255, type));
    }	
    return [nil, false];
  end
  

  def LangChecker.probeResponse(destination, probeType, responseType)
    probe = Scriptroute::ICMP.new(0);
    probe.ip_dst = destination;
    probe.icmp_type = probeType;
    probe.icmp_code = 0;
    probe.icmp_seq = 0;
    packets = Scriptroute::send_train([Struct::DelayedPacket.new(0, probe)]);
    return false if (!packets[0].response);
    response = packets[0].response.packet;
    if (response.is_a?(Scriptroute::ICMP) && 
	response.icmp_type == responseType) 
      return true;
    end
    return false;
  end
  
  def LangChecker.udp(destination)
    probe = Scriptroute::UDP.new(12) 
    probe.ip_dst = destination;
    begin 
      packets = Scriptroute::send_train([Struct::DelayedPacket.new(0, probe)]);
    rescue Exception => e
      puts "sluffing exception" + e
      return false
    end
    return false if (!packets[0].response);
    response = packets[0].response.packet;
    if (response.is_a?(Scriptroute::ICMP) &&
	response.icmp_type == Scriptroute::ICMP::ICMP_UNREACH &&
	response.icmp_code == Scriptroute::ICMP::ICMP_UNREACH_PORT )
      return true;
    end
    return false;
  end

  def LangChecker.tcp(destination)
    probe = Scriptroute::TCP.new(12) 
    probe.ip_dst = destination;
    packets = Scriptroute::send_train([Struct::DelayedPacket.new(0, probe)]);
    return false if (!packets[0].response);
    response = packets[0].response.packet;
    puts response
    return (response.is_a?(Scriptroute::TCP) && response.flag_rst)
  end

  def LangChecker.increasingIPIDs(destination, ttl, type)     
    #puts "lchecker: increasing ip-ids #{destination} #{ttl} #{type}";
    delayedArray = Array.new();
    (0..2).each { |i|
      probe = LangChecker.getDefaultPacket(type, i);
      probe.ip_dst = destination;
      probe.ip_ttl = ttl;
      delayedArray.push(Struct::DelayedPacket.new(0.001, probe));
    }
    begin
      packets = Scriptroute::send_train(delayedArray);
    rescue Exception => e
      $stderr.puts "Exception #{e} sending:"
      delayedArray.each_with_index { |dp,i| $stderr.puts "%d: %s" % [ i, dp.packet.to_s ] }
      return false
    end
    
    if (!packets[0] || !packets[1] || !packets[2] || 
                     !packets[0].response || !packets[1].response || !packets[2].response)
      return false;
    end
    
#      puts "%s (ttl=%d, %s): %s %s %s  --> %d %d %d" % [destination, ttl, type,
#        packets[0].response.packet.ip_src, 
#        packets[1].response.packet.ip_src, 
#        packets[2].response.packet.ip_src, 
#        packets[0].response.packet.ip_id,
#        packets[1].response.packet.ip_id,
#        packets[2].response.packet.ip_id];

#     puts "%.3f %.3f %.3f" % [packets[0].probe.time.to_f*1000, 
#      packets[1].probe.time.to_f*1000, 
#      packets[2].probe.time.to_f*1000];


    ##zeroes are being returned
    return false if (packets[0].response.packet.ip_id == 0 && packets[1].response.packet.ip_id == 0);
    
    ##the remote destination is copying my id's
    return false if (packets[0].response.packet.ip_id == packets[0].probe.packet.ip_id &&
		     packets[1].response.packet.ip_id == packets[1].probe.packet.ip_id);
    
    #puts "passed zero and copy test\n";

    #    packets.sort! { |a,b| a.probe.time <=> b.probe.time };
    ids = packets.map { |p| p.response.packet.ip_id };
    
    return true if (AliasResolution.before(ids[0]-10, ids[1]) && 
		    AliasResolution.before(ids[1], ids[0]+200) && 
		    ids[0] != ids[1] &&
		    AliasResolution.before(ids[1]-10, ids[2]) && 
		    AliasResolution.before(ids[2], ids[1]+200) && 
		    ids[1] != ids[2]);

    #puts "failed maternity test with #{ids.join(" ")}\n";

    return false;

  end

  def LangChecker.getDefaultPacket(type, seq=0) 
    if (type == "udp") 
      return Scriptroute::UDP.new(12);
    elsif (type == "tcp") 
      probe = Scriptroute::TCP.new(0);
      probe.th_dport=80;
      probe.th_win=1024;
      return probe;
    else 
      probe = if (type == "echo")
                Scriptroute::ICMPecho.new(0)
              elsif (type == "tstamp")
                Scriptroute::ICMPtstamp.new(0)
              else 
                raise "ERROR: unsupported packettype #{type} in sendAndReceiveTrain\n";
              end
      probe.icmp_seq=seq;
      return probe;
    end
  end

  def LangChecker.getBestPathOption4IPIDs(destination, ttl, router, tpath)
    canCheckForward = (LangChecker.increasingIPIDs(destination, ttl, "udp"))? 1 : -1;
    #puts "ccf = #{canCheckForward}\n";
    if (canCheckForward == -1) then
      ##todo: can be extended to other types, such as timestamps
      if (LangChecker.increasingIPIDs(router, 255, "echo")) then  
	subpath = TracePath.new(router);
	puts "subpathto #{router}:\n#{subpath.to_s}\n";
	if (tpath.subset(subpath)) then
	  return [router, 255, "echo", 1];	  
	end
      end
    end
    ## the default
    return [destination, ttl, "udp", canCheckForward];
  end



end ##module LangChecker


############### alias resolution: ally + ttl-testing + result cache #######
class AliasResolution 
  @@aliasCache = Hash.new(-1);
  @@logAllyResults = false;
  @@doTTLTesting = false;
  @@seqIPIDCache = Hash.new(-1);

  @@ally = (FileTest.executable?("/usr/bin/sr-ally"))? "/usr/bin/sr-ally" : "/usr/local/bin/sr-ally";
  
  def AliasResolution.cacheKey (ip1, ip2) 
    (ip1 < ip2)? "#{ip1}:#{ip2}" : "#{ip2}:#{ip1}";
  end

  def AliasResolution.aliases (hop1, hop2) 
    key = cacheKey(hop1.ip, hop2.ip);
    return @@aliasCache[key] if (@@aliasCache[key] != -1) 
      
    output = `#{@@ally} #{hop1.ip} #{hop2.ip}`
    output.gsub!(/\n/, " ");
    Kernel::system("echo '#{hop1.ip} #{hop2.ip} #{output}' >> ally.stats") if (@@logAllyResults);

    if (/ALIAS/.match(output) && !/NOT/.match(output)) then
      @@aliasCache[key] = true;
      return true;
    elsif (@@doTTLTesting and /UNKNOWN/.match(output)) then
      ttlSays = ttlBasedTest(hop1, hop2);
      Kernel::system("echo 'ttlTest: #{hop1} #{hop2} #{ttlSays}' >> ally.stats") if (@@logAllyResults);
      return ttlSays[0];
    else
      @@aliasCache[key] = false;
      return false;
    end
  end

  def AliasResolution.seqIPID (thop) 
    if (@@seqIPIDCache["#{thop.dst} #{thop.hop}"] == -1) 
      @@seqIPIDCache["#{thop.dst} #{thop.hop}"] = LangChecker.increasingIPIDs(thop.dst, thop.hop, "udp");
      puts "seqIPID results for #{thop.dst} #{thop.hop} is %s\n" % [@@seqIPIDCache["#{thop.dst} #{thop.hop}"]];
    end

    return @@seqIPIDCache["#{thop.dst} #{thop.hop}"];
  end


  def AliasResolution.ttlBasedTest (hop1, hop2) 
    return [false, "non-seq-ipid"] if (!seqIPID(hop1) or !seqIPID(hop2));
    
    a = Scriptroute::UDP.new(12) 
    b = Scriptroute::UDP.new(12) 
    a.ip_dst, a.ip_ttl = hop1.dst, hop1.hop;
    b.ip_dst, b.ip_ttl = hop2.dst, hop2.hop;
    
    packets = Scriptroute::send_train([ Struct::DelayedPacket.new(0,a), 
					Struct::DelayedPacket.new(0.001,b) ])
    
    return [false, "insufficient info"] if (!packets[0] || !packets[1] || 
					    !packets[0].response || !packets[1].response ||
					    packets[0].response.packet.ip_src != hop1.ip ||
					    packets[1].response.packet.ip_src != hop2.ip);
    
    id0 = packets[0].response.packet.ip_id
    id1 = packets[1].response.packet.ip_id
    return [false, "ids dont match"] if (!before(id0-10, id1) || !before(id1, id0+200) || id0 == id1);
    
    packetz = Scriptroute::send_train([ Struct::DelayedPacket.new(0,b), 
                                        Struct::DelayedPacket.new(0.001,a) ])
    
    return [false, "insufficient info2"] if (!packetz[0] || !packetz[1] || 
					     !packetz[0].response || !packetz[1].response ||
					     packetz[0].response.packet.ip_src != hop2.ip ||
					     packetz[1].response.packet.ip_src != hop1.ip);
    
    id2 = packets[0].response.packet.ip_id
    id3 = packets[1].response.packet.ip_id
    return [true, "excellent"] if(before(id2-10, id3) && before(id3, id2+200) && id3 != id2 &&
				  before(id0, id2) && before(id1, id3) && 
				  id2 != id1 && id2 != id0 && id3 != id1 && id3 != id0 );
    
    return [false, "final failure"];
  end
      
  def AliasResolution.before(seq1,seq2) 
    diff = seq1-seq2 
    # emulate signed short arithmetic.
    if (diff > 32767) then
      diff-=65535;
    elsif (diff < -32768) then
      diff+=65535;
    end
    # puts "#{seq1} #{diff < 0 ? "" : "not"} before #{seq2}\n"
    (diff < 0)
  end
end



############# generic train ################
class Train
  attr_reader :dsts, :numpackets, :resolution, :types, :ttls, :reverse_ttls, :num_losses, :packets, :shuffle;

  def getType (pkt) 
    if (pkt.is_a?(Scriptroute::TCP)) 
      return "tcp";
    elsif (pkt.is_a?(Scriptroute::UDP)) 
      return "udp";
    elsif (pkt.is_a?(Scriptroute::ICMP))
      if (pkt.icmp_type == Scriptroute::ICMP::ICMP_ECHO)
	return "echo";
      elsif (pkt.icmp_type == Scriptroute::ICMP::ICMP_TSTAMP)
	return "tstamp";
      end
    end
    raise "ERROR: unknown packet type: #{pkt.to_s}";
  end

  def getIndex (pr)
    pkt = pr.probe.packet;
    @dsts.each_index { |i|
      if (pkt.ip_dst == @dsts[i] and  
	  pkt.ip_ttl == @ttls[i] and 
	  getType(pkt) == @types[i]) 
	return i;
      end
    }
    raise "ERROR: could not classify packet: #{pkt.to_s}";
  end

  def initialize(dsts, numpackets, resolution, types, ttls, shuffle=false) 
    @dsts, @numpackets, @resolution, @types, @ttls, @shuffle = dsts, numpackets, resolution, types, ttls, shuffle;
    @reverse_ttls, @rem_srcs, @num_losses, @packets = Array.new(), Array.new(), Array.new(), Array.new();
    (@dsts).each_index { |i| 
      @packets[i], @reverse_ttls[i], @rem_srcs[i], @num_losses[i] = Array.new(),  Array.new(), Array.new(), 0;
    }

    num_dsts = @dsts.length;
    max_train = (MAX_TRAIN/num_dsts).floor;
    seq_id = 0; order = 0;
    allResponsePackets = Array.new();
    ( 0 .. (numpackets/max_train).ceil ).each { |trainnum|
      packetsThisRound = (numpackets < (trainnum+1)*max_train)? (numpackets - trainnum*max_train) : max_train;
      delayedPackets = Array.new();
      ( seq_id .. seq_id + packetsThisRound-1).each { |rep|
	
	##construct probe packets
	probes = Array.new();
	(@dsts).each_index { |i|
          case types[i]
          when "tcp"
	    probes[i] = Scriptroute::TCP.new(12);
          when "udp"
	    probes[i] = Scriptroute::UDP.new(12);
	    probes[i].uh_dport = 33444 ## unknown effect; five-tuple balancing may occur
          when "echo"
	    probes[i] = Scriptroute::ICMP.new(0)
            probes[i].icmp_type = Scriptroute::ICMP::ICMP_ECHO;
	    probes[i].icmp_code = 0;
	    probes[i].icmp_seq = rep;
          when "tstamp"
	    probes[i] = Scriptroute::ICMP.new(0)
            probes[i].icmp_type = Scriptroute::ICMP::ICMP_TSTAMP;
	    probes[i].icmp_code = 0;
	    probes[i].icmp_seq = rep;
          else 
            raise("ERROR: unsupported packettype #{types[i]} in sendAndReceiveTrain\n");
	  end	
	  probes[i].ip_ttl = @ttls[i];
	  probes[i].ip_dst = @dsts[i];
	  @dsts[i] = probes[i].ip_dst; # accomplishes a name lookup so responses can be classified - nspring

	}

	##insert probe packets in the order desired
	(@dsts).each_index { |i|
	  delay = (i==0)? resolution : 0;
	  #puts "delay: #{delay} (#{i})";
	  insert = (order + i) % num_dsts;
	  delayedPackets.push(Struct::DelayedPacket.new(delay, probes[insert]));
	  order = (order + 1) % num_dsts if (@shuffle);
	} 	  
      } ##each rep

      #puts "before real =  %.3f" % [Time.now.to_f*1000];
      allResponsePackets.push(Scriptroute::send_train(delayedPackets));
      #puts "after real =  %.3f" % [Time.now.to_f*1000];

    } ##each trainnum
    allResponsePackets.flatten!
    #allResponsePackets.sort!{ |a,b| a.probe.time <=> b.probe.time }

    allResponsePackets.each { |pr|
      if (pr and pr.probe and pr.probe.time) 
	i = getIndex(pr);
	@packets[i].push(pr);
	if (pr.response) then
	  @reverse_ttls[i].push(pr.response.packet.ip_ttl) if !@reverse_ttls[i].index(pr.response.packet.ip_ttl);
	  @rem_srcs[i].push(pr.response.packet.ip_src) if !@rem_srcs[i].index(pr.response.packet.ip_src);
	else
	  @num_losses[i] += 1;
	end
      end
    }

    # nspring thinks the following is a no-op.
    (@dsts).each_index { |i| 
      @packets[i].sort! { |a,b| a.probe.time <=> b.probe.time }
    }

    ##check sanity of the train
    (@dsts).each_index { |i|
      if (@reverse_ttls[i].length > 1) 
	puts "WARNING: too many reverse TTLs for index #{i}: #{@reverse_ttls[i].join(" ")}";
      end
      
      if (@rem_srcs[i].length > 1) 
	puts "WARNING: too many remote sources for index #{i}: #{@rem_srcs[i].join(" ")}";
      end

      ##todo: should really check a pattern in losses 
      ##      to distinguish between high loss rate and rate-limiting
      if (@num_losses[i] > 0.2*@numpackets) 
	#STDERR.puts "WARNING: too many losses #{@num_losses[i]} for index #{i}";
      end
    }

  end

  def to_s
    str = "train: #{@dsts} #{@numpackets} #{@resolution} (#{@types}) (#{@ttls})\n";
    @packets[0].each_index { |i| 
      (@dsts).each_index { |j|
	pr = @packets[j][i];
	stime = pr.probe.time.to_f * 1000;
	srcid = pr.probe.packet.ip_id;
	rtt = (pr.probe and pr.response) ? (pr.response.time - pr.probe.time) * 1000 : -1;
	str += "#{@dsts[j]} %d %.3f %d " %[rtt, stime, srcid];
	if (@types[j] == "tstamp") then 
	  rem_time =  (pr.response)? pr.response.packet.icmp_ttime : 0;
	  ott = (pr.response)? stime - pr.response.packet.icmp_ttime : -1;
	  str += "#{rem_time} %.3f " % [ott];
	end
      }
      str += "\n";
    }
    str += "losses: #{@num_losses.join(" ")}\n";
    return str;
  end

end

################ train2 ##############
class Train2
  attr_reader :dsts, :numpackets, :intratrain, :types, :ttls, :psizes, :numtrains, :intertrain, :reverse_ttls, :num_losses, :packets;

  def getProbe (i,j)
    dst = (@dsts[j])? @dsts[j] : @dsts[0];
    ttl = (@ttls[j])? @ttls[j] : @ttls[0];
    type = (@types[j])? @types[j] : @types[0];
    psize = (@psizes[j])? @psizes[j] : @psizes[0];
    
    probe = nil;
    case type
    when "udp"
      psize = (psize==-1)? 12 : psize;
      probe = Scriptroute::UDP.new(psize);
      probe.uh_dport = 33444;
    when "tcp"
      psize = (psize==-1)? 0 : psize;
      probe = Scriptroute::TCP.new(psize);
      probe.th_dport=80;
      probe.th_win=1024;
    else # "echo" or "tstamp"
      psize = (psize==-1)? 0 : psize;
      probe = Scriptroute::ICMP.new(psize);
      if (type == "echo")
	probe.icmp_type = Scriptroute::ICMP::ICMP_ECHO;
      elsif (type == "tstamp")
	probe.icmp_type = Scriptroute::ICMP::ICMP_TSTAMP;
      else 
	raise "ERROR: unsupported packettype #{i} #{j} #{type}\n";
      end
      probe.icmp_code = 0;
      probe.icmp_seq = i*numpackets + j;
    end	
    probe.ip_ttl = ttl;
    probe.ip_dst = dst;
    
    return probe;
  end


  def initialize(dsts, numpackets, intratrain, types, ttls, psizes, numtrains, intertrain) 
    @dsts, @numpackets, @intratrain, @types, @ttls, @psizes, @numtrains, @intertrain = 
      dsts, numpackets, intratrain,   types,  ttls,  psizes,  numtrains,  intertrain;
    @reverse_ttls, @rem_srcs, @num_losses, @packets = Hash.new(), Hash.new(), Array.new(), Array.new();

    ##forget the MAX_TRAIN complication for this part;

    delayedPackets = Array.new();
    (0..@numtrains-1).each { |i|
      delay = (i==0)? 0 : intertrain;
      (0..@numpackets-1).each { |j|
	probe = getProbe(i,j);
	#puts "#{i} #{delay}\n";
	delayedPackets.push(Struct::DelayedPacket.new(delay, probe));
	delay = intratrain;
	delay = (j==@numpackets-2)? 0.003 : 0.0005;
      }
    }
    @responsesPackets = Scriptroute::send_train(delayedPackets);

    ##check sanity of the train and insert it into @packets and @num_losses
    (0..@numtrains-1).each { |i|
      @packets[i] = Array.new();
      @num_losses[i] = 0;
      (0..@numpackets-1).each  { |j|
	index = i*@numpackets + j;
	pkt = @responsesPackets[index];
	@packets[i].push(pkt);

	dst = (@dsts[j])? @dsts[j] : @dsts[0];
	ttl = (@ttls[j])? @ttls[j] : @ttls[0];
	key = "#{dst}::#{ttl}";
	if (!@reverse_ttls[key]) 
	  @reverse_ttls[key] = Hash.new(0);
	  @rem_srcs[key] = Hash.new(0);
	end

	if (pkt) 
	  if (pkt.probe and pkt.probe.time)
	    if (pkt.response)
	      rttl = pkt.response.packet.ip_ttl;
	      src = pkt.response.packet.ip_src;
	      @reverse_ttls[key][rttl] += 1;
	      @rem_srcs[key][src] += 1;
	    else
	      @num_losses[i]+=1;
	    end
	    ##check the probe sort order
	    npkt = @responsesPackets[index+1];
	    if (npkt and npkt.probe and npkt.probe.time and pkt.probe.time > npkt.probe.time) 
	      STDERR.puts "ERROR: packets not in probe sorted order. \n%s %d %.3f\n%s %d %.3f\n" % 
		[pkt.probe.packet.ip_dst, pkt.probe.packet.ip_ttl, pkt.probe.time.to_f*1000, 
		npkt.probe.packet.ip_dst, npkt.probe.packet.ip_ttl, npkt.probe.time.to_f*1000];
	    end
	  end
	else
	  raise "ERROR: there are null responses\n";
	end
      }
    }
    
    @reverse_ttls.each_key { |key| 
      if (@reverse_ttls[key].keys.length > 1) 
	puts "WARNING: too many reverse TTLs for #{key}. #{@reverse_ttls[key].keys.join(" ")} #{@reverse_ttls[key].values.join(" ")} ";
      end
      
      if (@rem_srcs[key].keys.length > 1) 
	puts "WARNING: too many sources. #{@rem_srcs[key].keys.join(" ")} #{@rem_srcs[key].values.join(" ")} ";
      end
    }
	
  end

  def to_s
    str = "train: #{@dsts} #{@numpackets} #{@intratrain} (#{@types}) (#{@ttls})\n";
    (0..@numtrains-1).each { |i|
      str += "train #{i}\n";
      (0..@numpackets-1).each { |j| 
	pr = @packets[i][j];
	src = (pr.response)? pr.response.packet.ip_src : -1;
	srcid = pr.probe.packet.ip_id;
	stime = pr.probe.time.to_f * 1000;
	rtt = (pr.probe and pr.response) ? (pr.response.time - pr.probe.time) * 1000 : -1;
	ipid = (pr.response)? pr.response.packet.ip_id : -1;
	str += "#{src} %d %.3f %d " %[rtt, stime, ipid];
	if (@types[j] == "tstamp") then 
	  rem_time =  (pr.response)? pr.response.packet.icmp_ttime : 0;
	  str += "#{rem_time} ";
	end
      }
      str += "\n";
    }
    return str;
  end

  def simple_to_s
    str = "";
    (0..@numtrains-1).each { |i|
      (0..@numpackets-1).each { |j| 
	pr = @packets[i][j];
	src = (pr.response)? pr.response.packet.ip_src : -1;
	stime = pr.probe.time.to_f * 1000;
	ipid = (pr.response)? pr.response.packet.ip_id : -1;
	str += "%.3f %d " %[stime, ipid];
      }
      str += "\n";
    }
    return str;
  end

end


################## a traceroute hop ##################
class TraceHop
  @@aliasTesting = false;

  attr_reader :hop, :ip, :dst, :name;
  def initialize (hop, ip, dst, name=ip) 
    @hop = hop;
    @ip = ip;
    @dst = dst;
    @name = name;
  end
  
  ##todo: set this right
  def == (another)
    return true if (@ip == another.ip);
    return AliasResolution.aliases(self, another) if (@@aliasTesting);
    return false;
  end

  def to_s 
    return "#{@hop.to_s} #{@name.to_s} (#{@ip.to_s})";
  end

end

#################### the traceroute path ############

class TracePath
  attr_reader :path, :dst, :speaks_udp, :status_code;

  def initialize (dst, resolveNames=false, start_ttl=1, end_ttl=30)
    @path = Array.new;
    @dst = dst;
    @speaks_udp = (LangChecker.udp(@dst)) ? true : false;
    @status_code = "incomplete";

    #make self as first hop; simplifies binary search.
    @path.push(TraceHop.new(0, "0", dst));

    probe = Scriptroute::UDP.new(12)
    probe.ip_dst = dst
    port_unreach = false
  
    catch :port_unreachable do 
      (start_ttl..end_ttl).each { |ttl|
	probe.ip_ttl = ttl
	#puts "before trace: %.3f" % [Time.now.to_f];
	packets = Scriptroute::send_train([ Struct::DelayedPacket.new(0, probe) ]);
	#puts "after trace: %.3f" % [Time.now.to_f];
	if (!packets[0]) then 
	  STDERR.puts("WARNING: scriptrouted err'd. sleeping and retrying"); 
	  Kernel::sleep(5); 
	  redo; 
	end;
	response = (packets[0].response) ?  packets[0].response.packet : nil
	if (response.is_a?(Scriptroute::ICMP)) then
	  name = response.ip_src;
	  if (resolveNames) then
	    begin 
	      name = Socket.gethostbyname(response.ip_src)[0];
	    rescue 
	    end
	  end
	  @path.push(TraceHop.new(ttl, response.ip_src, dst, name)) if (!containsIP(response.ip_src));
	  port_unreach = true if (response.icmp_type == 3)
	  @status_code = "complete" if (response.icmp_type == 3 && response.icmp_code == 3) 
	else
	  port_unreach = true if (ttl - @path.last.hop >=2 and !@speaks_udp);
	end
	throw :port_unreachable if(port_unreach)
      }
    end
    
  end

  def containsIP(ip) 
    @path[1..-1].map { |p| return true if (p.ip == ip) };
    return false;
  end

  def getClosestTSNode(starthop, delta, minhop, maxhop) 
    while (starthop >= minhop && starthop <= maxhop) 
      if LangChecker.probeResponse(@path[starthop].ip, 
 				   Scriptroute::ICMP::ICMP_TSTAMP, 
 				   Scriptroute::ICMP::ICMP_TSTAMPREPLY) then
         return starthop;
       end
       starthop += delta;
     end
     return nil;
   end

   def subset (another, minhop = 1) 
     lasthop = another.path.last.hop;
     matches = Array.new();
     mismatches = Array.new();
     empties = Array.new();
     (minhop..lasthop).each { |hop|
       myhop = @path[hop];
       hishop = another.path[hop];
       if (myhop and hishop) 
	 (myhop == hishop)? matches.push(hop) : mismatches.push(hop);
       elsif (myhop or hishop) 
	 empties.push(hop);
       end
     }
     lastMatch = true;
     if (another.status_code == "incomplete")
       index = getIndex(another.dst);
       if (index < 0 or lasthop + 1 < index) then
	 lastMatch = false;
       end
     end

     #puts "subset results. matches: #{matches.join(" ")} mismatches: #{mismatches.join(" ")} empties: #{empties.join(" ")} lastMatch: #{lastMatch}";

     return true if ((mismatches.length == 0 or mismatches.last <= $prefixpath) && ##let go for local load balancing
		     lastMatch and matches.index(lasthop));   ##the last hop has to match
     return false;
   end

   def checkValidAnchors(farindex, farpath, nearindex, nearpath, verbose=true) 
     return (farpath.subset(nearpath) and subset(farpath, path[nearindex].hop));     
   end

   def getIndex(ip)
     (1..@path.length-1).each { |i|
       if (@path[i].ip == ip) 
	 return i;
       end
     }
     return -1;
   end

   def to_s 
     @path[1..-1].map { |p| p.to_s }.join("\n");
   end

end


end
end
